#include <avr/io.h>
#include <avr/interrupt.h>

#include <util/delay.h>

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <avr/eeprom.h>

#include "lcd.h"
#include "notes.h"

#define FULL_BOX 0xff
#define CURSOR 0x00
#define NEXT_ROW 0x30
#define ROW_UP 0
#define ROW_DOWN 1
#define CLEAR_BOX ' '
#define NOT_REVEALED 99
#define TOP_ROW 0x0f
#define BOTTOM_ROW 0x40
#define BTN_DELAY 250

typedef struct
{
	char symbol;
	char addr;
} tuple;

int song;
int start_song;
int stop_song;

uint8_t num_notes_mario = sizeof(mario_durations) / sizeof(mario_durations[0]);

uint8_t num_notes_bond = sizeof(bond_durations) / sizeof(bond_durations[0]);

void timer1_init()
{
    TCCR1A |= (2 << WGM10);
    TCCR1B |= (5 << CS10) | (3 << WGM12);
    TCCR1A |= (2 << COM1B0) | (2 << COM1A0);
    ICR1 = 60000;
    OCR1B = ICR1 / 4;
    OCR1A = ICR1 / 2;
}

void timer2_init()
{
    TCCR2A = (1 << WGM20) | (2 << COM2A0);
    TCCR2B = (6 << CS20);
    TIMSK2 = (1 << TOIE2);
    OCR2A = 0xC0;
}


void update_song()
{
    static uint8_t note_remaining = 0;
    static int8_t note_position = -1;
	
	if(stop_song != 1)
	{
		if(song == 1 && start_song == 1)
		{
			start_song = 0;
			note_remaining = 0;
			note_position = -1;
		}
		
		if(song == 0)
		{
			if(note_remaining-- == 0)
			{
				note_position = (note_position + 1) % num_notes_mario;
				int div = 1;
				if(num_notes_mario - note_position < 27)
					div = 1024;
				else
					div = 256;
				note_remaining = mario_durations[note_position] / 25;
				ICR1 = 16000000 / div / mario_notes[note_position];
				OCR1A = ICR1 / 2;
				TCNT1 = 0;
			}
		}
		
		else
		{
			if(note_remaining-- == 0)
			{
				note_position = (note_position + 1) % num_notes_mario;

				note_remaining = bond_durations[note_position] / 25;
				ICR1 = 16000000 / 1024 / bond_notes[note_position];
				OCR1A = ICR1 / 2;
				TCNT1 = 0;
			}		
		}
	}
}

ISR(TIMER2_OVF_vect)
{
    static uint8_t count = 0;

    count = (count + 1) % 3;
    if(count == 0)
        update_song();
}

void paintBox(uint8_t coord, char image)
{
	LCD_putCharAt(coord, image);
}

void clearScreen()
{
	int i;
	for(i = 0x00; i < 0x20; i++)
	{
		if(i < 0x10)
		{
			paintBox(i, CLEAR_BOX);
		}
		
		else
		{
			paintBox(i + NEXT_ROW, CLEAR_BOX);
		}
	}
}

void random_map(char map[32])
{
	int i = 0, aux, j = 0;
	for(i = 0; i < 32; i++)
	{
		j = rand() % 32;
		if(i != j)
		{
			aux = map[i];
			map[i] = map[j];
			map[j] = aux;
		}
	}
}

void init_map()
{
	int i = 0;
	for(i = 0x00; i < 0x20; i++)
	{
		if(i < 0x10)
		{
			paintBox(i, FULL_BOX);
		}
		
		else
		{
			paintBox(i + NEXT_ROW, FULL_BOX);
		}
	}
}

void hints(int leds[], int x)
{
	int i;

	PORTB = 0;
	PORTA &= ~(1 << PA5);
	for(i = 0; i < x; i++)
	{
		PORTB |= (1 << leds[i]);
	}
	
	if(x == 9)
	{
		PORTA |= (1 << PA5);
	}
}

void hot_or_cold(int leds[], char addr1, char addr2)
{
	int distance = 0;
	if((addr1 >= BOTTOM_ROW && addr2 >= BOTTOM_ROW) || (addr1 <= TOP_ROW && addr2 <= TOP_ROW))
	{
		distance = abs(addr1 - addr2);
	}
	
	else
	{
		if(addr1 >= BOTTOM_ROW)
			addr1 -= BOTTOM_ROW;
			
		if(addr2 >= BOTTOM_ROW)
			addr2 -= BOTTOM_ROW;
			
		distance = abs(addr1 - addr2) + 1;
	}
	
	switch(distance)
	{
		case 0:
			hints(leds, 9);
			break;

		case 1:
			hints(leds, 8);
			break;
			
		case 2:
			hints(leds, 7);
			break;
			
		case 3:
			hints(leds, 6);
			break;
			
		case 4:
			hints(leds, 5);
			break;
			
		case 5:
		case 6:
		case 7:
			hints(leds, 4);
			break;
			
		case 8:
		case 9:
		case 10:
			hints(leds, 3);
			break;
			
		case 11:
		case 12:
		case 13:
			hints(leds, 2);
			break;
			
		case 14:
		case 15:
		case 16:
			hints(leds, 1);
			break;
			
		default:
			hints(leds, 0);
			break;
	}
}


int main()
{

	DDRA &= ~(1 << PA4);
	DDRA &= ~(1 << PA3);
	DDRA &= ~(1 << PA2);
	DDRA &= ~(1 << PA1);
	DDRA &= ~(1 << PA0);
	PORTA |= (1 << PA0) | (1 << PA1) | (1 << PA2) | (1 << PA3) | (1 << PA4);
	DDRB |= ~0;
	DDRA |= (1 << PA5);
	DDRD &= ~(1 << PD6);
	DDRD |= (1 << PD5);
	PORTD |= (1 << PD6);
	
    uint32_t state;
    static uint32_t EEMEM sstate;
    state = eeprom_read_dword(&sstate);
    if (state == 0xffffffUL)
        state = 0xDEADBEEFUL;
    srand(state);
	eeprom_write_dword(&sstate, random());

    LCD_init();

	int leds[] = {PB0, PB1, PB2, PB3, PB4, PB7, PB5, PB6, PA5};
	uint8_t special_chars[8][8] = {
								{0x11,0xe,0x1e,0x1d,0x1b,0x1b,0x1f,0x1b},
								{0xe,0x1b,0x11,0x11,0x11,0x11,0x1f,0x1f},
								{0x4,0xe,0xe,0xe,0x1f,0x0,0x4,0x0},
								{0x2,0x3,0x2,0x2,0xe,0x1e,0xc,0x0},
								{0x0,0xe,0x15,0x17,0x11,0xe,0x0,0x0},
								{0x1,0x3,0xf,0xf,0xf,0x3,0x1,0x0},
								{0x0,0x0,0xa,0x1f,0x1f,0xe,0x4,0x0},
								{0x1f,0x15,0x1f,0x1b,0x1b,0x1f,0x11,0x1f}};
	
	LCD_writeInstr(LCD_INSTR_CGRAM);
	int i = 0, j = 0;
	for(i  = 0; i < 8; i++)
	{
		for(j = 0; j < 8; j++)
			LCD_writeData(special_chars[i][j]);
	}
	
	char symbols[16][3] = {{0x01, 0x00, 0x00}, {0x02, 0x00, 0x00}, {0x03, 0x00, 0x00}, {0x04, 0x00, 0x00},
							{0x05, 0x00, 0x00}, {0x06, 0x00, 0x00}, {0x07, 0x00, 0x00}, {'$', 0x00, 0x00},
							{0xf3, 0x00, 0x00}, {0xbc, 0x00, 0x00}, {0x7f, 0x00, 0x00}, {0x7e, 0x00, 0x00},
							{0xce, 0x00, 0x00}, {0xf7, 0x00, 0x00}, {'*', 0x00, 0x00}, {0xdb, 0x00, 0x00}};
							
	char hashsymbol[256];
	for(i = 0; i < 16; i++)
	{
		hashsymbol[(int)symbols[i][0]] = i;
	}
	
	
	LCD_writeInstr(LCD_INSTR_DDRAM);
	
	char permutation[32];
	char map[80][2];
	char current_addr;
	char current_row;
	char pairs_left = 16;
	tuple reveal1, reveal2;
	int moves;
	char moves_won[10];
	int running = 1;
	int cheat2 = 0;
	int song_on = 1;
	//running = 0;
	
    timer1_init();
    timer2_init();
	sei();

	while(running)
	{
		for(i = 0x00; i < 0x20; i++)
		{
			if(i < 0x10)
			{
				permutation[(int)i] = i;
			}
		
			else
			{
				permutation[(int)i] = i + 0x30;
			}
		}
	
		random_map(permutation);
	
		j = 0;
		for(i = 0; i < 16; i++)
		{
			symbols[i][1] = permutation[j++];
			symbols[i][2] = permutation[j++];
			map[(int)symbols[i][1]][0] = symbols[i][0];
			map[(int)symbols[i][2]][0] = symbols[i][0];
			map[(int)symbols[i][1]][1] = FULL_BOX;
			map[(int)symbols[i][2]][1] = FULL_BOX;		
		
		}
	
	
		init_map();
		current_addr = 0x00;
		current_row = ROW_UP;
		paintBox(current_addr, CURSOR);
		pairs_left = 16;
	
		reveal1.symbol = NOT_REVEALED;
		reveal1.addr = NOT_REVEALED;
		reveal2.symbol = NOT_REVEALED;
		reveal2.addr = NOT_REVEALED;
		moves = 0;
	
		while(1)
		{
			if((PIND & (1 << PD6)) == 0)
			{
				if(song_on == 1)
				{
					cli();
					PORTD &= ~(1 << PD5);
					song_on = 0;
					stop_song = 1;
				}
				
				else
				{
					sei();
					song_on = 1;
					stop_song = 0;
				}
			}
			
			/* win */
			if(pairs_left == 0)
				break;
			
			/* cheat */
			if((PINA & (1 << PA2)) == 0 && (PINA & (1 << PA1)) == 0)
			{
				clearScreen();
				break;
			}
			
			
			/* cheat 2 */
			if((PINA & (1 << PA0)) == 0 && (PINA & (1 << PA4)) == 0)
			{
				if(cheat2 == 0)
				{
					clearScreen();
					for(i = 0x00; i < 0x20; i++)
					{
						if(i < 0x10)
						{
							if(map[i][1] != CLEAR_BOX)
								paintBox(i, map[i][0]);
						}
					
						else
						{
							if(map[i + 0x30][1] != CLEAR_BOX)
								paintBox(i + 0x30, map[i + 0x30][0]);
						}
					}
					
					cheat2 = 1;
					_delay_ms(BTN_DELAY);
				}
				
				else
				{
					clearScreen();
					for(i = 0x00; i < 0x20; i++)
					{
						if(i < 0x10)
						{
							paintBox(i, map[i][1]);
						}
					
						else
						{
							paintBox(i + 0x30, map[i + 0x30][1]);
						}
					}
					
					cheat2 = 0;
					_delay_ms(BTN_DELAY);
				}
			}
				
			/* move RIGHT */
			if((PINA & (1 << PA0)) == 0)
			{
				if(current_row == ROW_UP && current_addr < 0x0f)
				{
					if(cheat2 == 1)
						paintBox(current_addr, map[(int)current_addr][0]);
					else
						paintBox(current_addr, map[(int)current_addr][1]);
					current_addr++;
					paintBox(current_addr, CURSOR);
				}
			
				else
				if(current_row == ROW_DOWN && current_addr < 0x4f)
				{
					if(cheat2 == 1)
						paintBox(current_addr, map[(int)current_addr][0]);
					else
						paintBox(current_addr, map[(int)current_addr][1]);
					current_addr++;
					paintBox(current_addr, CURSOR);
				}
				_delay_ms(BTN_DELAY);
			}
			
			/* move UP */
			if((PINA & (1 << PA1)) == 0)
			{
				if(current_row == ROW_DOWN)
				{
					if(cheat2 == 1)
						paintBox(current_addr, map[(int)current_addr][0]);
					else
						paintBox(current_addr, map[(int)current_addr][1]);
					current_addr -= 0x40;
					paintBox(current_addr, CURSOR);
					current_row = ROW_UP;
				}
				_delay_ms(BTN_DELAY);
			}
			
			/* move DOWN */
			if((PINA & (1 << PA2)) == 0)
			{		
				if(current_row == ROW_UP)
				{
					if(cheat2 == 1)
						paintBox(current_addr, map[(int)current_addr][0]);
					else
						paintBox(current_addr, map[(int)current_addr][1]);
					current_addr += 0x40;
					paintBox(current_addr, CURSOR);
					current_row = ROW_DOWN;
				}
				_delay_ms(BTN_DELAY);
			}
			
			/* action CENTERBUTTON */
			if((PINA & (1 << PA3)) == 0)
			{
				/* we reveal the first tile */
				if(reveal1.symbol == NOT_REVEALED && map[(int)current_addr][1] != CLEAR_BOX)
				{
					map[(int)current_addr][1] = map[(int)current_addr][0];
					paintBox(current_addr, map[(int)current_addr][0]);
					reveal1.symbol = map[(int)current_addr][0];
					reveal1.addr = current_addr;
					moves++;
				}
				
				/* we reveal the second tile */
				if(reveal1.symbol != NOT_REVEALED && map[(int)current_addr][1] != CLEAR_BOX
					&& reveal1.addr != current_addr)
				{
					paintBox(current_addr, map[(int)current_addr][0]);
					_delay_ms(1000);
					
					if(reveal1.symbol != map[(int)current_addr][0])
					{
						map[(int)reveal1.addr][1] = FULL_BOX;
						paintBox(current_addr, map[(int)current_addr][1]);
						paintBox(reveal1.addr, FULL_BOX);
						char addr_other_symbol;
						int sym_id = hashsymbol[(int)reveal1.symbol];
						if(symbols[sym_id][1] != reveal1.addr)
						{
							addr_other_symbol = symbols[sym_id][1];
						}
						
						else
						{
							addr_other_symbol = symbols[sym_id][2];
						}
												
						hot_or_cold(leds, current_addr, addr_other_symbol);
					}
					
					else
					{
						map[(int)reveal1.addr][1] = map[(int)reveal1.addr][0] = CLEAR_BOX;
						map[(int)current_addr][1] = map[(int)current_addr][0] = CLEAR_BOX;
						paintBox(current_addr, CLEAR_BOX);
						paintBox(reveal1.addr, CLEAR_BOX);
						pairs_left--;
						hints(leds, 9);
					}
					
					reveal1.symbol = NOT_REVEALED;
					reveal1.addr = NOT_REVEALED;
					moves++;
					paintBox(current_addr, CURSOR);
				}
				_delay_ms(BTN_DELAY);
			}

			/* move LEFT */
			if((PINA & (1 << PA4)) == 0)
			{
				if(current_row == ROW_UP && current_addr > 0x00)
				{
					if(cheat2 == 1)
						paintBox(current_addr, map[(int)current_addr][0]);
					else
						paintBox(current_addr, map[(int)current_addr][1]);
					current_addr--;
					paintBox(current_addr, CURSOR);
				}
				
				else
				if(current_row == ROW_DOWN && current_addr > 0x40)
				{
					if(cheat2 == 1)
						paintBox(current_addr, map[(int)current_addr][0]);
					else
						paintBox(current_addr, map[(int)current_addr][1]);
					current_addr--;
					paintBox(current_addr, CURSOR);
				}
				_delay_ms(BTN_DELAY);
			}
		}
		
		clearScreen();
		LCD_printAt(0x00, "Won in ");
		sprintf(moves_won, "%d ", moves);
		LCD_printAt(0x07, moves_won);
		LCD_print("moves");
		LCD_printAt(0x44, "Continue?");
		LCD_printAt(0x40, "YES");
		LCD_printAt(0x4E, "NO");
		
		/* Choose yes or no */
		while((PINA & (1 << PA4)) != 0)
		{
			if((PINA & (1 << PA0)) == 0)
			{
				running = 0;
				break;
			}
		}
	}
	
	/* Credits */
	clearScreen();
	song = 1;
	start_song = 1;
	LCD_printAt(0x00, "Created by");
	_delay_ms(2500);
	LCD_printAt(0x40, "Adrian Dobrica");
	_delay_ms(2500);
	LCD_printAt(0x00, "               ");
	_delay_ms(2500);
	LCD_printAt(0x00, "331 CC");
	_delay_ms(2500);
	LCD_print("-ACS");
	_delay_ms(2500);
	LCD_print("-UPB");
	_delay_ms(2500);
	LCD_printAt(0x40, "               ");
	_delay_ms(2500);
	LCD_printAt(0x40, "PM 2015");
	_delay_ms(2500);
	LCD_printAt(0x00, "               ");
	_delay_ms(2500);
	LCD_printAt(0x00, "               ");
	_delay_ms(2500);
	LCD_printAt(0x00, "Thanks for");
	_delay_ms(500);
	LCD_printAt(0x40, "               ");
	LCD_printAt(0x48, "playing!");

	while(1);
    return 0;
}
